/*------------------------------------------------------------------------/
/  Experimental SDv2 (in SD mode) control module
/-------------------------------------------------------------------------/
/
/  Copyright (C) 2024, Evan Hillas, all right reserved.
/
/ * This software is a free software and there is NO WARRANTY.
/ * No restriction on use. You can use, modify and redistribute it for
/   personal, non-profit or commercial products UNDER YOUR RESPONSIBILITY.
/ * Redistributions of source code must retain the above copyright notice.
/
/-------------------------------------------------------------------------/
    v0.1, 4 Oct 2024

  Features and Limitations:

  * Platform Specific - Prop2
    Uses hardware DMA resources and "smartpin"s.

  * No Media Change Detection
    Application program needs to perform a f_mount() after media change.

/-------------------------------------------------------------------------*/

#include "diskio.h"    // Common include file for FatFs and disk I/O layer

#ifndef __propeller2__
#error ONLY SUPPORTS THE PROPELLER 2 MICROCONTROLLER
#endif
#include <stdlib.h>    // needed for some data types
#include <propeller2.h>


//--------------------------------------------------------------------------
// Compile options
//--------------------------------------------------------------------------

#ifndef NOREADBLOCKCRC
#define SD_BLOCK_CRC    // enables block read CRC checking
#endif
//#define PR_PARAMS
//#define SD_DEBUG


enum {
    RX_SCHMITT = 0,    // 0 = CMOS threshold (CLK pin is Schmitt), 1 = Schmitt trigger of CMD and DAT pins
    CMDAT_REG = 0,    // 0 = unregistered pins, 1 = registered CMD and DAT pins
    CLK_REG = 0,    // 0 = unregistered pin, 1 = registered CLK pin
    CLK_POL = 0,    // 0 = idle low (selects High-Speed), 1 = idle high (selects Default-Speed)
//    CLK_DIV = 3,    // minimum of 2 (SD clock divider of sysclock), 3 when BLOCK_CRC defined

    COGCRCBUFF = 126,    // cogRAM address of CRC processing buffer (130 longwords size)
};

/*
typedef enum {    // diskio.h
	RES_OK = 0,		// 0: Successful
	RES_ERROR,		// 1: R/W Error
	RES_WRPRT,		// 2: Write Protected
	RES_NOTRDY,		// 3: Not Ready
	RES_PARERR		// 4: Invalid Parameter
} DRESULT;

// Disk Status Bits (DSTATUS)  -  diskio.h
#define STA_NOINIT		0x01	// Drive not initialized
#define STA_NODISK		0x02	// No medium in the drive
#define STA_PROTECT		0x04	// Write protected
*/


//--------------------------------------------------------------------------
//
//   Module Private Functions
//
//--------------------------------------------------------------------------


// Private allocation of cogRAM parameter passing registers, instead of the public PRx registers
extern register uint32_t 
    *r_param1, *r_param2, *r_param3, *r_param4,
    *r_param5, *r_param6, *r_param7, *r_param8;

#define  r_clk      r_param1
#define  r_cmd      r_param2
#define  r_leadin   r_param3
#define  r_nco      r_param4
#define  r_se2      r_param5
#define  r_bdelay   r_param6
#define  r_ca       r_param7

#define  r_timeout  r_param5
#define  r_resp     r_param6
#define  r_clkr     r_param7
#define  r_clks     r_param8

#define  r_pdat     r_param2
#define  r_mdat     r_param6
#define  r_clkdiv   r_param7
#define  r_mclk     r_param8

#define  r_crc      r_param7
#define  r_leadin2  r_param8



// SD Card/Socket specific variables
uint8_t  cid[18];
uint16_t  rca16;    // active card's 16-bit RCA register
uint16_t  clkdivider;    // full speed clock divider
uint32_t  expectedblk;
uint8_t  writing;
uint8_t  rxlagcomp;    // Compensates for cumulative clk-rx latencies,
                                // calibration triggers on read errors
uint8_t  clkpin;
uint8_t  cmdpin;
uint8_t  dat0pin;
uint8_t  powpin;
uint8_t  ledpin;

DSTATUS  Stat;    // Disk status




//-----------------------------------------------------------------------
//-----------------------------------------------------------------------
static void  releasepins( void )
{
    int  PIN_CLK = clkpin;
    int  PIN_CMD = cmdpin;
    int  PIN_DAT = dat0pin | 3<<6;
    int  PIN_POW = powpin;
    int  PIN_LED = ledpin;

    _pinf( PIN_CLK );
    _pinf( PIN_CMD );
    _pinf( PIN_DAT );
    _pinf( PIN_POW );
    _pinf( PIN_LED );
    _wrpin( PIN_CLK, 0 );
    _wrpin( PIN_CMD, 0 );
    _wrpin( PIN_DAT, 0 );
    _wrpin( PIN_POW, 0 );
    _wrpin( PIN_LED, 0 );
}



//-----------------------------------------------------------------------
// Check if SD card is inserted in socket
//-----------------------------------------------------------------------

static int  sdcard_present( void )    // 0:Card Missing, 1:Card Present
{
    int  PIN_POW = powpin;

    _pinf(PIN_POW);
    _wrpin(PIN_POW, P_SCHMITT_A);
    _waitus(1);    // wait for pull-up resistor to act if there is no card
    if( !_pinr(PIN_POW) )    // An inserted card will pull this down against the 22k ohm pull-up
        return 1;

#ifdef SD_DEBUG
    __builtin_printf(" SD Card not detected\n");
#endif

    return 0;
}



//-----------------------------------------------------------------------
//-----------------------------------------------------------------------

static int  sdcard_power( void )
{
    int  PIN_VOLT = clkpin;
    int  PIN_CMD = cmdpin;
    int  PIN_DAT = dat0pin | 3<<6;
    int  PIN_POW = powpin;
    uint32_t  samples, threshold;
    int32_t  tmr, timeout;

    releasepins();    // SD spec 6.4.2
    if( !sdcard_present() )
        return 0;
#ifdef SD_DEBUG
    __builtin_printf(" Card detected ... power cycle of SD card\n");
#endif

// Use CompDAC pin mode to wait for supply to fall below 0.5 Volt, SD spec 6.4.1.2
    threshold = (int)(0.5 / 3.4 * 255.0);    // 0.5 Volts
#ifdef SD_DEBUG
    __builtin_printf("  power-down threshold = %d   pin state = %d\n", threshold, _pinr(PIN_VOLT));
#endif
    _wrpin(PIN_VOLT, P_LEVEL_A | threshold<<8);    // set CompDAC threshold to 0.5 volts
    _pinh(PIN_POW);    // power off the SD card by overpowering the 1800 ohm Card Detect
    tmr = _cnt();
    _pinl(PIN_DAT);    // drive DAT pin group low to discharge the now unpowered capacitors
    _pinl(PIN_CMD);    // ditto CMD
    _waitus(1);    // DAC settle time is about 1.0 us
    samples = _pinr(PIN_VOLT);
    do {
        _waitus(32);    // 32 us x 32 samples = 1.0 ms
        samples = samples<<1 | _pinr(PIN_VOLT);
    } while( samples );    // unanimous wait for pin to drop below threshold level
    tmr = _cnt() - tmr;

#ifdef SD_DEBUG
    tmr = _muldiv64(tmr, 1_000_000, _clockfreq());
    __builtin_printf("  power-down slope = %d us   pin state = %d\n", tmr, _pinr(PIN_VOLT));
#endif

    // Use CompDAC pin mode to wait for supply to rise above 2.7 Volts, SD spec 6.4.1.3
    threshold = (int)(2.7 / 3.3 * 255.0 + 0.5);    // 2.7 Volts
    timeout = 36 * 32;    // 35 * 32 * 31 us = 35 ms, + 1 ms to cover filter effect
#ifdef SD_DEBUG
    __builtin_printf("  power-up threshold = %d   pin state = %d\n", threshold, _pinr(PIN_VOLT));
#endif
    _pinf(PIN_DAT);
    _pinf(PIN_CMD);
    _pinf(PIN_POW);    // and power back on, released back to Card Detect function
    tmr = _cnt();
    _wrpin(PIN_VOLT, P_LEVEL_A | threshold<<8);    // set CompDAC threshold to 2.7 volts
    _waitus(1);    // DAC settle time is about 1.0 us
    samples = !_pinr(PIN_VOLT);
    do {
        _waitus(32);    // 32 us x 32 samples = 1.0 ms
        samples = samples<<1 | !_pinr(PIN_VOLT);
        timeout--;
    } while( samples && timeout );    // unanimous wait for pin to rise above threshold level
    tmr = _cnt() - tmr;
    if( samples ) {
#ifdef SD_DEBUG
        __builtin_printf("  error:  Volts missing, at SD power-up\n");
#endif
        return 0;
    }
#ifdef SD_DEBUG
    tmr = _muldiv64(tmr, 1_000_000, _clockfreq());
    __builtin_printf("  power-up slope = %d us   pin state = %d\n", tmr, _pinr(PIN_VOLT));
#endif
    releasepins();

    return 1;
}



//-----------------------------------------------------------------------
//-----------------------------------------------------------------------

static uint32_t  crc7sd(    // SD spec 4.5
    uint8_t *buf,
    size_t len )
{
    uint32_t  crc = 0;
    uint32_t  val;

    __asm {
// Reference code courtesy of Ariba
crc7lp
		rdlong	val, buf
		add	buf, #4
		movbyts	val, #0b00_01_10_11    // byte swap within longwords
		setq	val
		crcnib	crc, #0x48    // CCITT 7-bit polynomial is x0 + x3 + x7
		crcnib	crc, #0x48    // 0x09 reversed and shifted for CRCNIB
		djz	len, #crc7done
		crcnib	crc, #0x48
		crcnib	crc, #0x48
		djz	len, #crc7done
		crcnib	crc, #0x48
		crcnib	crc, #0x48
		djz	len, #crc7done
		crcnib	crc, #0x48
		crcnib	crc, #0x48
		djnz	len, #crc7lp
crc7done
		rev	crc         // correct the bit order to match standard
		shr	crc, #24
		or	crc, #1    // add the SD response end-bit as 8th bit
    }
    return crc;
}



//-----------------------------------------------------------------------
// Tx data block's DMA parameter setup
//-----------------------------------------------------------------------

static struct txblk_parms_t {
    uint32_t  p_clk, p_dat, m_leadin, m_nco, m_se2, m_dat, m_crc, m_leadin2;
} txblkset;

#ifdef PR_PARAMS
enum {
      r_clk = 0x1e0, r_pdat, r_leadin, r_nco, r_se2, r_mdat, r_crc, r_leadin2
};
#endif


static void  set_txblkset(
    uint32_t clkdiv )
{
    int  PIN_DAT0 = dat0pin;

    txblkset.p_clk = clkpin;
    txblkset.p_dat = PIN_DAT0 | 3<<6;
    txblkset.m_se2 = 0b110<<6 | PIN_DAT0;    // trigger on high level - card ready
    txblkset.m_leadin = X_IMM_32X1_1DAC1 | 5 + CLK_REG - CMDAT_REG + clkdiv + (CLK_POL ? clkdiv : clkdiv/2);  // start(S)-bit is ahead of the first streamer bit
    txblkset.m_leadin2 = X_IMM_32X1_1DAC1 | 5 + CLK_REG - CMDAT_REG + (CLK_POL ? clkdiv : clkdiv/2);
    txblkset.m_nco = 0x8000_0000UL / clkdiv + (0x8000_0000UL % clkdiv > 0 ? 1 : 0);  // data transfer rate, round up upon a non-zero remainder
    txblkset.m_dat = X_RFBYTE_4P_1DAC4 | PIN_DAT0<<17 | X_PINS_ON | X_ALT_ON | 512 * 2;  // mode and nibble count
    txblkset.m_crc = X_IMM_8X4_1DAC4 | PIN_DAT0<<17 | X_PINS_ON | 8;  // mode and nibble count
}


//-----------------------------------------------------------------------
// Transmit a data block to the card
//-----------------------------------------------------------------------

static int  tx_datablock(    // $3F = no response, $00 = busy timeout, %10_101_1 = CRC mismatch, %10_010_1 = SD card happy
    const uint8_t *buf,    // Buffer address of block data to be transmitted
    uint32_t timeout )    // 250 ms timeout for card busy, in sysclock ticks
{
    __asm const {    // "const" enforces XIP, "volatile" enforces Fcache
// retrieve parameter set
		loc	pa, #txblkset    // fast copy from struct
		setq	#sizeof(txblkset) / 4 - 1
		rdlong	r_clk, pa    // fast copy to cogRAM
// start clocking to see when card exits from write busy status
		wypin	##-1, r_clk    // busy check
		getct	pb    // timeout begins
		add	timeout, pb    // SDHC 250 ms timeout of block erase-write (SD spec fig 4-35)

		setse2	r_se2    // trigger on high level - card ready
#ifdef SD_DEBUG
//		loc	ptrb, #dbuf+8    // dbuf[2]
#endif
// preload block data into CogRAM for CRC processing
		setq	#512/4-1    // one SD block + CRC
		rdlong	COGCRCBUFF, buf    // fast copy to cogRAM before engaging the FIFO

		pollse1    // clear stale event
		pollse2    // clear stale event
    }

    __asm volatile {    // "const" enforces XIP, "volatile" enforces Fcache
		rdfast	lnco, buf    // start the FIFO
		setxfrq	lnco
		setq	timeout    // apply the 250 ms timeout to WAITSEn
		waitse2   wz    // wait for ready (DAT0 high), Z set if timed-out - still busy
		wypin	#2, r_clk    // busy turnaround, Nwr (SD spec fig 4-32)
#ifdef SD_DEBUG
//		getct	pa    // latency measurement
//		sub	pa, pb
//		wrlong	pa, ptrb++    // store the latency interval
#endif
		waitse1

		dirl	r_clk    // reset smartpin
	if_nz	drvl	r_pdat    // start-bit placed on data pins

	if_nz	xinit	r_leadin, #0    // lead-in delay from here at sysclock/1
	if_nz	setq	r_nco        // streamer transfer rate (takes effect with buffered command below)
	if_nz	xzero	r_mdat, #0     // rx buffered-op, aligned with clock via lead-in
		dirh	r_clk    // clock timing starts here
	if_nz	wypin	clocks, r_clk    // first pulse outputs during second clock period

// verify the CRC
	if_nz	rep	@.rend1, #512/4    // one SD block + CRC
	if_nz	alti	ptr, #0b100_111_000    // next D-field substitution, then increment ptr, result goes to PA
	if_nz	movbyts	pa, #0b00_01_10_11    // byte swap within longword
	if_nz	splitb	pa    // 8-nibble order swapped to bit order of 4 bytes
	if_nz	setq	pa
	if_nz	crcnib	crc3, poly
	if_nz	crcnib	crc3, poly
	if_nz	crcnib	crc2, poly
	if_nz	crcnib	crc2, poly
	if_nz	crcnib	crc1, poly
	if_nz	crcnib	crc1, poly
	if_nz	crcnib	crc0, poly
	if_nz	crcnib	crc0, poly
.rend1
	if_nz	rolbyte	pa, crc3, #0
	if_nz	rolbyte	pa, crc2, #0
	if_nz	rolbyte	pa, crc1, #0
	if_nz	rolbyte	pa, crc0, #0
	if_nz	mergeb	pa    // 4-byte order swapped to bit order of 8 nibbles
#ifdef SD_DEBUG
//	if_nz	wrlong	pa, ptrb++
#endif
	if_nz	waitse1

	if_nz	setxfrq	lnco
	if_nz	dirl	r_clk    // reset smartpin
	if_nz	xinit	r_leadin2, #0    // lead-in delay from here at sysclock/1
	if_nz	setq	r_nco        // streamer transfer rate (takes effect with buffered command below)
	if_nz	xzero	r_crc, pa    // 8 nibbles, transmit low nibble first
	if_nz	dirh	r_clk    // clock timing starts here
	if_nz	wypin	#16+2+CLK_POL, r_clk    // 16-bit CRC + end-bit
	if_nz	rolbyte	pa, crc3, #1
	if_nz	rolbyte	pa, crc2, #1
	if_nz	rolbyte	pa, crc1, #1
	if_nz	rolbyte	pa, crc0, #1
	if_nz	mergeb	pa    // 4-byte order swapped to bit order of 8 nibbles
	if_nz	xzero	r_crc, pa    // 8 nibbles, transmit low nibble first
#ifdef SD_DEBUG
//	if_nz	wrlong	pa, ptrb++
#endif
	if_nz	xzero	r_crc, endbit
	if_nz	waitse1

// tx complete, now rx "CRC status"
		dirl	r_pdat    // release DAT pins ready for CRC status from SD card
		and	r_pdat, #0x3f    // strip away ADDPINS, leaving DAT0

		rep	@.rend2, #6
		wypin	#1, r_clk    // one clock pulse
		testp	r_pdat   wc    // sample DAT0 before the clock pin transitions
		rcl	timeout, #1    // collect status bits
		waitse1      // wait for each clock done
.rend2
		ret

clocks		long 1 + 512 * 2    // start-bit + nibble count
lnco		long 0x8000_0000UL    // tells RDFAST/WRFAST not to wait for FIFO ready
endbit		long 0xffff_ffffUL
poly		long 0x8408    // CRCNIB shifts right, therefore ordered from left: x0 + x5 + x12 + x16
ptr		long 0x1f6<<19 | COGCRCBUFF<<9    // register PA for ALTI result substitution, and CRC buffer address
crc3		long 0
crc2		long 0
crc1		long 0
crc0		long 0
    }
    __asm {
		wypin	#500, r_clk    // trailing clocks
		zerox	timeout, #5    // remove excess bits from "CRC status"
    }

#ifdef SD_DEBUG
    if( timeout == 0x37 )
        __builtin_printf(" BlockWriteError: No CRC response\n");
    if( timeout == 0b10_101_1 )
        __builtin_printf(" BlockWriteError: CRC mismatch response\n");
    if( timeout == 0 )
        __builtin_printf(" BlockWriteError: Data busy timeout\n");
#endif

    return timeout;    // $3F = no response, $00 = busy timeout, %10_101_1 = CRC mismatch, %10_010_1 = SD card happy
}



//-----------------------------------------------------------------------
// Rx data block's DMA parameter setup
//-----------------------------------------------------------------------

static struct rxblk_parms_t {
    uint32_t  p_clk, p_dat, m_leadin, m_nco, m_se2, m_dat, m_clkdiv, m_clks;
} rxblkset;

#ifdef PR_PARAMS
enum {
      r_clk = 0x1e0, r_pdat, r_leadin, r_nco, r_se2, r_mdat, r_clkdiv, r_clks
};
#endif


static void  set_rxblkset(
    uint32_t clkdiv,
    uint32_t rxlag )
{
    int  PIN_DAT0 = dat0pin;

    rxblkset.p_clk = clkpin;
    rxblkset.p_dat = PIN_DAT0 | 3<<6;
    rxblkset.m_se2 = 0b100<<6 | PIN_DAT0+3;    // trigger on low level, preferred to falling edge trigger because
        // DAT pins idle high during command and response and also ensures best chance of seeing an early start-bit
    if( clkdiv < 20 )  {
        rxblkset.m_leadin = X_IMM_32X1_1DAC1 | 8 + clkdiv + rxlag;    // skip past the start(S)-bit
        rxblkset.m_clks = 10 | 5<<16;    // start-bit search clock divider
    } else {
        rxblkset.m_leadin = X_IMM_32X1_1DAC1 | 8 + clkdiv*2 + rxlag;    // skip past the start(S)-bit
        rxblkset.m_clks = clkdiv | clkdiv/2<<16;    // start-bit search clock divider
    }
    rxblkset.m_clkdiv = clkdiv | clkdiv/2<<16;
    rxblkset.m_nco = 0x8000_0000UL / clkdiv + (0x8000_0000UL % clkdiv > 0 ? 1 : 0);  // data transfer rate, round up upon a non-zero remainder
#ifdef SD_BLOCK_CRC
    rxblkset.m_dat = X_4P_1DAC4_WFBYTE | PIN_DAT0<<17 | X_PINS_ON | X_ALT_ON | 512 * 2 + 16;  // mode and nibble count
#else
    rxblkset.m_dat = X_4P_1DAC4_WFBYTE | PIN_DAT0<<17 | X_PINS_ON | X_ALT_ON | 512 * 2;  // mode and nibble count
#endif
}


//-----------------------------------------------------------------------
// Receive data blocks from the card
//-----------------------------------------------------------------------

static int  rx_datablocks(    // Remaining block count, 0 = Success
    uint8_t *buf,    // Buffer address to store the received data blocks
    uint32_t blocks,    // Number of data blocks to be read
    uint32_t timeout,    // 100 ms timeout for each start-bit latency, in sysclock ticks
    uint8_t *crcbuf )    // extra hubRAM for holding the final 16 CRC nibbles
{
    uint32_t  crc3, crc2, crc1, crc0;

    __asm const {    // "const" enforces XIP, "volatile" enforces Fcache
		loc	pa, #rxblkset    // fast copy from struct
		setq	#sizeof(rxblkset) / 4 - 1
		rdlong	r_clk, pa    // fast copy to registers PR8..PR15

		fltl	r_pdat
		setse2	r_se2    // trigger on low level, preferred to falling edge trigger
		// because DAT pins idle high during command and response and also ensures best chance
		// of seeing an early start-bit
		pollse1
    }

    __asm volatile {    // "const" enforces XIP, "volatile" enforces Fcache
#ifdef SD_BLOCK_CRC
		modz	_clr   wz    // clear Z flag to create a one-shot for first block
nextblk
	if_z	setq	#512/4+16/8-1    // one SD block + CRC
	if_z	rdlong	COGCRCBUFF, buf    // fast copy to cogRAM before engaging the FIFO
	if_z	add	buf, blkoffset
		wrfast	lnco, buf    // setup FIFO for streamer use
// no room in the buffer for final CRC-16 nibbles, to be handled later
		cmp	blocks, #2   wc    // last block?
	if_c	sub	clocks, #16+2    // stop the clocks before the CRC-16 nibbles
	if_c	sub	r_mdat, #16    // truncate DMA to end of buffer
#else
		modz	_clr   wz    // clear Z flag to create a one-shot for first block
		wrfast	lnco, buf    // setup FIFO for streamer use
nextblk
#endif
// locate data block start-bit
// NOTE: Search code has a one SD clock lag.  The start(S)-bit has been and gone by the time it triggers.
//   The following streamer setup begins its sampling ahead of the second data bit arriving.
		wxpin	r_clks, r_clk    // clock divider for start-bit search
		wypin	lnco, r_clk    // start the search
		getct	pb    // timeout begins
		pollse2    // clear stale event
		setxfrq	lnco        // set streamer to sysclock/1 for lead-in timing
		add	pb, timeout    // SDHC 100 ms timeout of start-bit search, Nac (SD spec 4.12.4 note 1)
		setq	pb    // apply the 100 ms timeout to WAITSEn
		waitse2   wc    // wait for DAT3 falling edge, a timeout sets C
		dirl	r_clk    // halt clock gen ASAP
		wxpin	r_clkdiv, r_clk    // clock divider for data block

// start-bit found, now read the data block
	if_nc	xinit	r_leadin, #0    // lead-in delay from here at sysclock/1
	if_nc	setq	r_nco        // streamer transfer rate (takes effect with buffered command below)
	if_nc	xzero	r_mdat, #0     // rx buffered-op, aligned with clock via lead-in
		dirh	r_clk    // clock timing starts here
	if_nc	wypin	clocks, r_clk    // first pulse outputs during second clock period

#ifdef SD_BLOCK_CRC
	if_nc	call	#crc_check
// prep for next block+CRC
	if_nc	waitse1    // wait for clocking to complete
if_nc_and_z	djnz	blocks, #nextblk

// now collect final block CRC-16 nibbles
if_nc_and_z	wrfast	lnco, crcbuf    // setup FIFO for streamer use, assumes mdat holds stale streamer mode
if_nc_and_z	setxfrq	lnco    // set streamer to sysclock/1 for lead-in timing
if_nc_and_z	dirl	r_clk    // reset clock smartpin
if_nc_and_z	setword	r_mdat, #16, #0    // reuse streamer mode for collecting the CRC-16 nibbles

if_nc_and_z	xinit	r_leadin, #0    // lead-in delay from here at sysclock/1
if_nc_and_z	setq	r_nco        // streamer transfer rate (takes effect with buffered command below)
if_nc_and_z	xzero	r_mdat, #0     // rx buffered-op, aligned with clock via lead-in
if_nc_and_z	dirh	r_clk    // clock timing starts here
if_nc_and_z	wypin	#16+2, r_clk    // CRC-16 nibbles + end bit

#ifdef SD_DEBUG
		mov	timeout, #0
	if_c	mov	timeout, #3    // timed out
#endif
if_c_or_nz	ret    // exit on error
		waitse1    // wait for clocking to complete

// copy block and CRC to cogRAM
		setq	#512/4-1    // 512 bytes, one SD block, 4 bytes per longword
		rdlong	COGCRCBUFF, buf    // fast copy to cogRAM
		setq	#16/8-1    // 16 CRC-16 nibbles, 8 nibbles per longword
		rdlong	COGCRCBUFF+512/4, crcbuf
// do the check
		call	#crc_check
	if_nz	add	blocks, #1    // indicate a failed final CRC
#else
	if_nc	waitse1    // wait for clocking to complete
	if_nc	djnz	blocks, #nextblk
#ifdef SD_DEBUG
		mov	timeout, #0
	if_c	mov	timeout, #3    // timed out
#endif
#endif
		ret

lnco		long 0x8000_0000UL    // lead-in rate, sysclock/1
clocks		long 512 * 2 + 16 + 2    // clock pulses, nibble count + 16-bit CRC + end-bit
#ifdef SD_BLOCK_CRC
poly		long 0x8408    // CRCNIB shifts right, therefore ordered from left: x0 + x5 + x12 + x16
cogaddr		long 0x1f6<<19 | COGCRCBUFF<<9    // register PA for ALTI result substitution, and CRC buffer address
blkoffset	long 512

crc_check
		mov	crc3, #0
		mov	crc2, #0
		mov	crc1, #0
		mov	crc0, #0
		mov	pb, cogaddr
// check CRC of prior read SD block
	if_z	rep	@.rend, #512/4+16/8    // one SD block + CRC
	if_z	alti	pb, #0b100_111_000    // next D-field substitution, then increment PB, result goes to PA
	if_z	movbyts	pa, #0b00_01_10_11    // byte swap within longword
	if_z	splitb	pa    // 8-nibble order swapped to bit order of 4 bytes
	if_z	setq	pa
	if_z	crcnib	crc3, poly
	if_z	crcnib	crc3, poly
	if_z	crcnib	crc2, poly
	if_z	crcnib	crc2, poly
	if_z	crcnib	crc1, poly
	if_z	crcnib	crc1, poly
	if_z	crcnib	crc0, poly
	if_z	crcnib	crc0, poly
.rend
// pass/fail
		or	crc3, crc2
		or	crc1, crc0
		or	crc3, crc1   wz    // Z set for pass, clear for fail
		long 0xfd64002d    // RET
#endif
    }

#ifdef SD_DEBUG
    if( blocks ) {
        __builtin_printf(" BlockReadError:  blocks=%d  err=%d  crc=%08x  ", blocks, timeout, crc3);
        if( timeout == 3 )
            __builtin_printf("start-bit timed out\n");
        else
            __builtin_printf("CRC failed\n");
    }
#endif

    return blocks;    // number of blocks remaining, 0 = Success
}



//-----------------------------------------------------------------------
// Command response's DMA parameter setup
//-----------------------------------------------------------------------

static struct cmdresp_parms_t {
    uint32_t  p_clk, p_cmd, m_leadin, m_nco, timeout, m_resp, m_clkr, m_clks;
} crset;

#ifdef PR_PARAMS
enum {
      r_clk = 0x1e0, r_cmd, r_leadin, r_nco, r_timeout, r_resp, r_clkr, r_clks
};
#endif


static void  set_respset(
    uint32_t clkdiv,
    uint32_t rxlag )
{
    int  PIN_CMD = cmdpin;

    crset.p_clk = clkpin;
    crset.p_cmd = PIN_CMD;
    crset.m_resp = X_1P_1DAC1_WFBYTE | PIN_CMD<<17 | X_PINS_ON | X_ALT_ON;    // + bit count
    crset.m_nco = 0x8000_0000UL / clkdiv + (0x8000_0000UL % clkdiv > 0 ? 1 : 0);  // data transfer rate, round up upon a non-zero remainder
    if( clkdiv < 20 )  {    // start-bit search ends on T-bit instead of S-bit
        crset.m_leadin = X_IMM_32X1_1DAC1 | 8 + rxlag;    // store start(S)-bit ahead of clock pulse
        crset.timeout = 10 * 64;    // max of 64 clocks for response to start, Ncr (SD spec 4.12.4)
        crset.m_clks = 10 | 5<<16;    // start-bit search clock divider
    } else {
        crset.m_leadin = X_IMM_32X1_1DAC1 | 8 + clkdiv + rxlag;    // store start(S)-bit ahead of clock pulse
        crset.timeout = clkdiv * 64;    // max of 64 clocks for response to start, Ncr (SD spec 4.12.4)
        crset.m_clks = clkdiv | clkdiv/2<<16;    // start-bit search clock divider
    }
    crset.m_clkr = clkdiv | clkdiv/2<<16;    // response clock divider
}


//-----------------------------------------------------------------------
// Receive the card's command-response packet
//-----------------------------------------------------------------------

static int  rx_response(    // 0:Start-bit timed out
    uint8_t *resp,    // Buffer address for response bytes
    uint32_t len,    // Number of bytes in response, including the start-end framing bits
    uint32_t clocks )    // Additional clocks for card's post-response idle transition
{
    uint32_t lnco = 0x8000_0000UL;

    __asm const {    // "const" enforces XIP, "volatile" enforces Fcache
		loc	pb, #crset    // fast copy from struct
		setq	#sizeof(crset) / 4 - 1
		rdlong	r_clk, pb    // fast copy to registers PR8..PR15

		shl	len, #3
		setword	r_resp, len, #0    // add bit count to DMA mode word, crset.m_resp
		add	clocks, len
		mov	pb, #0b010<<6    // trigger on falling edge of CMD pin - start-bit arrived
		or	pb, r_cmd
//		setse2	#0    // cancel triggering before reuse - not needed if POLLSE2 is used
		setse2	pb
    }

    __asm volatile {    // "const" enforces XIP, "volatile" enforces Fcache
		waitse1    // wait for end of command tx clocks

// locate response start-bit
// NOTE: Search code has a one SD clock lag.  The start(S)-bit has been and gone by the time it triggers.
//   The following streamer setup samples the T bit twice before further data bits arrive.
		outh	r_cmd    // when driven, momentarily force CMD pin high before start-bit search
		wxpin	r_clks, r_clk    // clock divider for start-bit search
		wypin	lnco, r_clk    // start the search
		fltl	r_cmd    // pin drive release before SD card takes over
		pollse2    // clear stale event
		getct	pb       // timeout begins
		wrfast	lnco, resp    // engage the FIFO
		setxfrq	lnco    // set sysclock/1 for lead-in timing
		add	pb, r_timeout    // max of 64 clocks for response to start, Ncr (SD spec 4.12.4)
		setq	pb    // apply the 64 clock timeout to WAITSEn
		waitse2   wc    // wait for CMD start bit, C set if timed-out
		dirl	r_clk    // halt clock gen ASAP
		wxpin	r_clkr, r_clk    // revert to regular clock gen

// start-bit found, now read the response (NOTE: start-bit is part of first byte)
	if_nc	xinit	r_leadin, #0    // lead-in delay from here at sysclock/1
	if_nc	setq	r_nco    // streamer transfer rate (takes effect with buffered command below)
	if_nc	xzero	r_resp, #0     // rx buffered-op aligned with clock via lead-in
		dirh	r_clk    // clock timing starts here
	if_nc	wypin	clocks, r_clk    // first clock pulse outputs during second clock period

		// twiddle thumbs
	if_c	mov	len, #0    // start-bit timed out

	if_nc	waitxfi    // wait for rx data completion before resuming hubexec
    }
#ifdef SD_DEBUG
    if( !len )
        __builtin_printf(" Response start-bit timeout\n");
#endif

    return len;    // 0:Start-bit timed out
}



//-----------------------------------------------------------------------
// Command's DMA parameter setup
//-----------------------------------------------------------------------

static struct cmd_parms_t {
    uint32_t  p_clk, p_cmd, m_leadin, m_nco, m_se2, busydelay, m_ca;
} cmdset;

#ifdef PR_PARAMS
enum {
      r_clk = 0x1e0, r_cmd, r_leadin, r_nco, r_se2,  r_bdelay, r_ca
};
#endif


static void  set_cmdset(
    uint32_t clkdiv )
{
    int  PIN_CMD = cmdpin;
    int32_t  bdelay = (int32_t)clkdiv * 8 - 54;

    cmdset.p_clk = clkpin;
    cmdset.p_cmd = PIN_CMD;
    cmdset.m_se2 = 0b100<<6 | dat0pin;    // trigger on low level - card busy
//    cmdset.m_se2 = 0b110<<6 | dat0pin;    // trigger on high level - card ready
    cmdset.busydelay = bdelay < 0 ? 0 : bdelay;
    cmdset.m_leadin = X_IMM_32X1_1DAC1 | 5 + CLK_REG - CMDAT_REG + (CLK_POL ? clkdiv : clkdiv/2);
    cmdset.m_nco = 0x8000_0000UL / clkdiv + (0x8000_0000UL % clkdiv > 0 ? 1 : 0);  // data transfer rate, round up upon a non-zero remainder
    cmdset.m_ca = X_IMM_32X1_1DAC1 | PIN_CMD<<17 | X_PINS_ON | 32;    // mode and bit count
}


//-----------------------------------------------------------------------
// Transmit a command packet to the card
//-----------------------------------------------------------------------

static int  tx_command(    // 0:CMD pin stuck low
    uint32_t cmd,    // Command
    uint32_t arg )    // Argument
{
    __asm const {    // "const" enforces XIP, "volatile" enforces Fcache
		loc	pb, #cmdset    // fast copy from struct
		setq	#sizeof(cmdset) / 4 - 1
		rdlong	r_clk, pb    // fast copy to cogRAM

		setse1	#0    // cancel triggering before reuse - not needed if POLLSE1 is used
		mov	pb, r_clk
		or	pb, #0b001<<6    // trigger on rising edge - clocks completed
		wypin	#1, r_clk    // ensure card releases the CMD bus, don't use #1 !!!
		setse1	pb    // must be immediately after WYPIN to catch IN rising
//		setse1	#0b110<<6 | PIN_CLK    // trigger on high level - smartpin finished

		or	cmd, #0x40    // add T bit in front of command
		shl	cmd, #24    // command at [31:24] is needed for later CRC
		mov	pb, arg
		shr	pb, #8
		or	pb, cmd
		rev	pb    // first 32 bits prep'd for streamer
		setxfrq	##0x8000_0000UL    // set sysclock/1 for lead-in timing
#ifdef SD_DEBUG
//		loc	ptrb, #dbuf
//		wrlong	pb, ptrb++
#endif
		waitse1    // wait for clock pulse to complete
		fltl	r_clk    // reset clock gen smartpin, needed to instruction align the samrtpin cycle
		testp	r_cmd   wc    // send new command only if CMD pin has returned high
// C is set when CMD pin is high
	if_c	drvl	r_cmd    // drive CMD pin low, start-bit, ready for streamer output

// begin command phase
	if_c	xinit	r_leadin, #0    // lead-in delay at sysclock/1
	if_c	setq	r_nco    // data transfer rate, takes effect on XZERO below (buffered command)
	if_c	xzero	r_ca, pb    // place first 32 bits in tx buffer, aligned with clock via lead-in delay
		dirh	r_clk    // clock timing starts here, first clock pulse occurs in smartpin's second period
		wypin	#6*8+1, r_clk    // SD clock pulses, 6 bytes + 1 pulse to kick off a response

// CRC7 algorithm, SD spec 4.5
	if_c	setq	cmd    // first byte, CMD<<24
	if_c	mov	pb, #0
	if_c	crcnib	pb, #0x48    // CCITT polynomial is x7 + x3 + 1
	if_c	crcnib	pb, #0x48    // 0x09 reversed and shifted for CRCNIB
	if_c	setq	arg     // + four bytes
	if_c	crcnib	pb, #0x48
	if_c	crcnib	pb, #0x48
	if_c	crcnib	pb, #0x48
	if_c	crcnib	pb, #0x48
	if_c	crcnib	pb, #0x48
	if_c	crcnib	pb, #0x48
	if_c	crcnib	pb, #0x48
	if_c	crcnib	pb, #0x48

	if_c	or	pb, #0x180    // add end-bit after the CRC
	if_c	rev	arg
	if_c	rolbyte	pb, arg, #3    // insert last byte of arg ahead of CRC

	if_c	xzero	r_ca, pb     // place remaining 16 bits in buffer for when first 32 bits completes
#ifdef SD_DEBUG
//		wrword	pb, ptrb++
#endif
	if_nc	mov	cmd, #0    // CMD pin stuck low - Probably needs power down
    }
#ifdef SD_DEBUG
    if( !cmd )
        __builtin_printf(" CMD pin stuck low\n");
#endif

    return cmd;    // 0:CMD pin stuck low
}


#if 0
static int  send_cmd(    // 0:Okay, 2:CMD pin low, 3:Response timed-out
    uint32_t cmd,    // Command
    uint32_t arg,    // Argument
    uint8_t *resp    // Response Buffer.  If resp == NULL then no response expected
)
{
    int  rc;

    __asm const {    // "const" enforces XIP, "volatile" enforces Fcache
		loc	pa, #cmdset    // fast copy from struct
		setq	#sizeof(cmdset) / 4 - 1
		rdlong	r_clk, pa    // fast copy to registers PR8..PR15

#ifdef CLK_STEP
		wypin	#16, r_clk    // ensure CMD bus is released
#else
		wypin	#8, r_clk    // ensure CMD bus is released
#endif
		mov	pa, r_clk
		or	pa, #0b001<<6    // trigger on rising edge - clocks completed
		setse1	#0    // cancel triggering before reuse - not needed if POLLSE1 is used
		setse1	pa
//		setse1	#0b110<<6 | PIN_CLK    // trigger on high level - smartpin finished

		or	cmd, #0x40    // add T bit in front of command
		shl	cmd, #24    // command at [31:24] is needed for later CRC
		mov	pb, arg
		shr	pb, #8
		or	pb, cmd
		rev	pb    // first 32 bits prep'd for streamer
		setxfrq	##0x8000_0000UL    // set sysclock/1 for lead-in timing
#ifdef _DIAG
		loc	ptrb, #dbuf
		wrlong	pb, ptrb++
#endif
		waitse1    // wait for clocks to complete
		testp	r_cmd   wc    // send new command only if CMD pin has returned high
// C is set when CMD pin is high
	if_c	drvl	r_cmd    // drive CMD pin low, start-bit, ready for streamer output
		fltl	r_clk    // reset clock gen smartpin, needed to instruction align the samrtpin cycle

// begin command phase
	if_c	xinit	r_leadin, #0    // lead-in delay at sysclock/1
	if_c	setq	r_nco    // data transfer rate, takes effect on XZERO below (buffered command)
	if_c	xzero	r_ca, pb    // place first 32 bits in tx buffer, aligned with clock via lead-in delay
		dirh	r_clk    // clock timing starts here, first clock pulse occurs in smartpin's second period
#ifdef CLK_STEP
		wypin	#6*16+2, r_clk    // SD clock steps, 6 bytes + 1 tick to kick off a response
#else
		wypin	#6*8+1, r_clk    // SD clock pulses, 6 bytes + 1 tick to kick off a response
#endif
		mov	rc, #0
	if_nc	mov	rc, #2    // error code 2 if CMD pin stayed low - Probably needs power down

// CRC7 algorithm, SD spec 4.5
	if_c	setq	cmd    // first byte, CMD<<24
		mov	pb, #0    // return a zero value CRC if SD card is busy
	if_c	crcnib	pb, #0x48    // CCITT polynomial is x7 + x3 + 1
	if_c	crcnib	pb, #0x48    // 0x09 reversed and shifted for CRCNIB
	if_c	setq	arg     // + four bytes
	if_c	crcnib	pb, #0x48
	if_c	crcnib	pb, #0x48
	if_c	crcnib	pb, #0x48
	if_c	crcnib	pb, #0x48
	if_c	crcnib	pb, #0x48
	if_c	crcnib	pb, #0x48
	if_c	crcnib	pb, #0x48
	if_c	crcnib	pb, #0x48

	if_c	or	pb, #0x180    // add end-bit after the CRC
	if_c	rev	arg
	if_c	rolbyte	pb, arg, #3    // insert last byte of arg ahead of CRC

	if_c	xzero	r_ca, pb     // place remaining 16 bits in buffer for when first 32 bits completes
#ifdef _DIAG
		wrword	pb, ptrb++
#endif
    }

    if( !rc && resp )  {
        switch( cmd )  {
        case 2:    // ALL_SEND_CID
        case 9:    // SEND_CSD
        case 10:    // SEND_CID
            if( rx_response(17, resp) )    // R2 responses, 17-bytes!
                return 0;
            rc = 3;
            break;
/*
        case 7:
            if( arg )  {    // SELECT_CARD
                if( rx_response(6, resp) )    // R1b response
                    return 0;
                rc = 3;
                break;
            }    // DESELECT_CARD    // no response expected
        case 0:    // GO_IDLE_STATE    // no response expected
        case 4:    // SET_DSR    // no response expected
        case 15:    // GO_INACTIVE_STATE    // no response expected
            arg = clkpin;
            __asm {
		waitse1    // wait for end of command tx clocks
		wypin	#80, arg    // trailing clocks
            }
            return 0;
*/
        default:
            if( rx_response(6, resp) )    // R1, R1b, R3, R6, R7 responses
                return 0;
            rc = 3;
            break;
        }
    } else {    // either CMD pin stayed low, or resp == NULL for no response expected
        pb = clkpin;
        __asm {
		waitse1    // wait for end of command tx clocks
		wypin	#80, pb    // trailing clocks
        }
    }
#ifdef SD_DEBUG
    if( rc == 1 )
        puts(" Card-Busy (DAT0 low) timed out!");
    if( rc == 2 )
        puts(" CMD pin stuck low!");
    if( rc == 3 )
        puts(" Command response missing!");
#endif
    return rc;
}
#endif



//-----------------------------------------------------------------------
// Setup the SD clock pin. Also configure DMA parameter sets
//-----------------------------------------------------------------------

static void  sdcard_clockconf(
    uint32_t clkdiv )
{
    int  PIN_CLK = clkpin;
    int  PIN_CMD = cmdpin;
    int  PIN_DAT = dat0pin | 3<<6;
    int  PIN_POW = powpin;
    uint32_t  pinmode, rxlag = rxlagcomp;

    // update the setsets
    set_cmdset(clkdiv);
    set_respset(clkdiv, rxlag);
    set_rxblkset(clkdiv, rxlag);
    set_txblkset(clkdiv);

    // configure pins
    pinmode = (CMDAT_REG ? P_SYNC_IO : 0) | (RX_SCHMITT ? P_SCHMITT_A : 0);
    _wrpin(PIN_CMD, pinmode);
    _wrpin(PIN_DAT, pinmode);
    _wrpin(PIN_POW, P_SCHMITT_A);

    // SD clock gen smartpin
    pinmode = P_PULSE | P_OE | (CLK_POL ? P_INVERT_OUTPUT : 0) | P_SCHMITT_A | (CLK_REG ? P_SYNC_IO : 0);
    _pinstart(PIN_CLK, pinmode, clkdiv | clkdiv/2<<16, 0);

#ifdef SD_DEBUG
    __builtin_printf(" SD clock divider set to sysclock/%d.  'rxlag' compensation is %d\n", clkdiv, rxlag);
#endif
}



//-----------------------------------------------------------------------
// Send a common housekeeping command packet to the card
//-----------------------------------------------------------------------

static int  send_cmd(    // 0:Fail, 1:Success
    uint32_t cmd,    // Command
    uint32_t arg,    // Argument
    uint8_t *resp )    // Buffer address for response bytes
{
#ifdef SD_DEBUG
//    __builtin_printf(" CMD%d - ", cmd);
#endif
    tx_command(cmd, arg);
    if( !resp ) {    // expecting no response
        int  PIN_CLK = clkpin;
        int  PIN_CMD = cmdpin;
        __asm {
		waitse1    // wait for end of command tx clocks
		outh	PIN_CMD    // momentarily force CMD pin high before new clocks
		wypin	#8, PIN_CLK    // trailing clocks
		fltl	PIN_CMD    // pin drive release before SD card takes over
		waitse1    // wait for end of command tx clocks
		wypin	#500, PIN_CLK    // trailing clocks
        }
        return 1;
    }
    if( rx_response(resp, 6, -500) )    // R1,R1b,R3,R6,R7 responses
        if( resp[5] == crc7sd(resp, 5) )
            return 1;
#ifdef SD_DEBUG
    __builtin_printf(" CMD%d error!\n", cmd);
#endif
    return 0;
}



//-----------------------------------------------------------------------
// Send a R2 housekeeping command packet to the card
//-----------------------------------------------------------------------

static int  send_cmd_r2(    // 0:Fail, 1:Success
    uint32_t cmd,    // Command
    uint32_t arg,    // Argument
    uint8_t *resp )    // Buffer address for response bytes
{
#ifdef SD_DEBUG
//    __builtin_printf(" CMD%d - ", cmd);
#endif
    tx_command(cmd, arg);
    if( rx_response(resp, 17, -500) )    // R2 response, response length, trailing clocks
        if( resp[16] == crc7sd(&resp[1], 15) )
            return 1;
#ifdef SD_DEBUG
    __builtin_printf(" CMD%d error!\n", cmd);
#endif
    return 0;
}



//-----------------------------------------------------------------------
// Send an application specific command packet to the card
//-----------------------------------------------------------------------

static int  send_acmd(    // 0:Fail, 1:Success
    uint32_t acmd,    // App command
    uint32_t arg,    // Argument
    uint8_t *resp )    // Buffer address for response bytes
{
    if( send_cmd(55, rca16<<16, resp) ) {    // ACMD prefix
#ifdef SD_DEBUG
//    __builtin_printf(" ACMD%d - ", acmd);
#endif
        tx_command(acmd, arg);
        if( rx_response(resp, 6, -500) )
            return 1;    // skip CRC processing, not all ACMDs have one
    }
#ifdef SD_DEBUG
    __builtin_printf(" ACMD%d error!\n", acmd);
#endif
    return 0;
}



//-----------------------------------------------------------------------
// Wait for card ready - Checks the DAT0 pin write-data-busy indicator
//-----------------------------------------------------------------------

static int  wait_ready(    // 0:Card Busy, 1:Card Ready
    uint32_t timeout )    // 100/250/500 ms interval in sysclock ticks
{
    uint32_t  m_se2;
    int  ready;

    _wypin(clkpin, -1);    // lots of clocks
    timeout += _cnt();    // start the timer
    m_se2 = 0b110<<6 | dat0pin;    // trigger on high level - card ready
    ready = 0;   // Busy

    __asm {    // "const" enforces XIP, "volatile" enforces Fcache
		setse2	#0    // cancel triggering before reuse - not needed if POLLSE2 is used
		setse2	m_se2
		setq	timeout
		waitse2   wc
	if_nc	mov	ready, #1    // Not busy
    }
#ifdef SD_DEBUG
    if( !ready )
        __builtin_printf(" BUSY-TIMEOUT ");
#endif

    return ready;
}



//-----------------------------------------------------------------------
//-----------------------------------------------------------------------

static uint32_t  calibrate_rxlag(    // 0:Failed calibration,  1..16:New rxlag
    uint8_t *resp,    // buffer address for response bytes
    uint8_t *cidbytes )    // address of stored copy of the card's CID register
{
    uint32_t  rca = rca16 << 16, CLK_DIV = clkdivider;
    uint32_t  rxlag, repeats, highest = 0, lowest = 0;

    send_cmd(7, 0, NULL);    // DESELECT, no response expected

    for( rxlag = 1; rxlag <= 16; rxlag++ )
    {
        set_respset(CLK_DIV, rxlag);

        repeats = 80;
        do {
#ifdef SD_DEBUG
            putchar('.');
#endif
            if( !send_cmd_r2(10, rca, resp) )    // CMD10  SEND_CID - R2 response
                break;
            if( memcmp(cidbytes, resp, 17) )
                break;
        } while( --repeats );

        if( !repeats ) {    // pass
            highest = rxlag;
            if( !lowest )
                lowest = rxlag;
        } else
            if( highest )    // done
                break;
#ifdef SD_DEBUG
        puts("");
#endif
    }

    if( highest )  {
        rxlag = lowest + (highest - lowest) / 2;
//        rxlag = rxlag > lowest ? rxlag : highest;
        rxlagcomp = rxlag;

        set_respset(CLK_DIV, rxlag);
        set_rxblkset(CLK_DIV, rxlag);
    } else
        rxlag = 0;
#ifdef SD_DEBUG
    __builtin_printf( "  rxlag=%d selected  Lowest=%d Highest=%d\n", rxlag, lowest, highest );
#endif
    send_cmd(7, rca, resp);    // CMD7  SELECT_CARD - "standby" to "tran" state, R1b response

    return rxlag;
}



//--------------------------------------------------------------------------
//
//   Public Functions
//
//--------------------------------------------------------------------------


//-----------------------------------------------------------------------
// Get Disk Status                                                       
//-----------------------------------------------------------------------

DSTATUS disk_status(
    BYTE drv        // Drive number (always 0)
)
{
    if( drv )
        return STA_NOINIT;

    return Stat;
}



//-----------------------------------------------------------------------
// Initialize Disk Drive
//-----------------------------------------------------------------------
enum {
    CMD8CHECKPAT = 0x100|0x5a  // 3V3 bit and check pattern, SD spec 4.3.13
};


DSTATUS disk_initialize (
    BYTE drv )        // Physical drive nmuber (0)
{
    int  PIN_CLK = clkpin;
    uint32_t  rca, tmr, CLK_DIV = clkdivider;
    uint8_t  *resp = __builtin_alloca(20);    // response buffer, in hubRAM
    DSTATUS  status;

    status = STA_NOINIT | STA_NODISK;
    Stat = status;
    writing = 0;
    expectedblk = -1;

    if( drv ) {
#ifdef SD_DEBUG
        __builtin_printf("bad drv %d\n", drv);
#endif
        return status;
    }

    sdcard_power();    // attempt to power cycle the SD card, ignore outcome

// setup SD clock smartpin, and other pins, and fill out initial setsets
    rca = (_clockfreq() + 200_000UL) / 400_000UL;    // divider for 400 kHz SD clock
    rxlagcomp = 0 + CLK_REG + CMDAT_REG;
    sdcard_clockconf(rca);

// give it a kick
    _wypin(PIN_CLK, 500);    // SD spec 6.4.1
    _waitms(1);
    send_cmd(0, 0, NULL);    // CMD0  GO_IDLE_STATE - in case the power cycle didn't happen
    send_cmd(8, CMD8CHECKPAT, resp);    // CMD8  SEND_IF_COND - 3.3V and check pattern, SD spec 4.3.13, R7 response
    _waitms(1);

// start conversation with SD card
    send_cmd(0, 0, NULL);    // CMD0  GO_IDLE_STATE - no response expected
    if( !send_cmd(8, CMD8CHECKPAT, resp) )    // CMD8  SEND_IF_COND - 3.3V and check pattern, SD spec 4.3.13, R7 response
        goto faillabel;    // no response, probably card missing
#ifdef SD_DEBUG
    __builtin_printf("Card idle OK\n");
#endif
    status = STA_NOINIT;

// confirms SD v2.0+, with valid voltage and check pattern
    if( __builtin_bswap32(*(long *)&resp[1]) != CMD8CHECKPAT )
        goto faillabel;

// verify the card is SDHC or better
    rca = 0;
    rca16 = 0;    // "idle" state uses RCA=0 in ACMDs
    tmr = _getms();
    do {    // host is HCS capable, 3.3 Volt, and requests XPC (full power)
        if( send_acmd(41, 1<<30 | 1<<20 | 1<<28, resp) ) {   // ACMD41  SD_SEND_OP_COND - "idle" to "ready" state, R3 response
            rca = __builtin_bswap32(*(long *)&resp[1]);
//            rca = resp[4];
//            if( rca>>7 )    // valid Ready bit, card has switched from "idle" to "ready" state
            if( rca>>31 )    // valid Ready bit, card has switched from "idle" to "ready" state
                break;
        }
        _wypin(PIN_CLK, 400);
        _waitms(1);    // card busy, polling less than 50 ms, SD spec 4.4
    } while( _getms() - tmr < 1000 );    // SD spec 4.2.3, 1.0 sec timeout, probably gone "inactive"
#ifdef SD_DEBUG
    __builtin_printf(" OCR register %08x  - ",rca);
#endif

    if( rca>>30 != 3 )    // confirm CCS bit
//    if( rca>>6 != 3 )    // confirm CCS bit
        goto faillabel;

#ifdef SD_DEBUG
    __builtin_printf("Valid SDHC/SDXC Card\n");
#endif

//-----------------------------------------
// CMD11 voltage switch not used
//   Default 3.3V is ideal for Prop2 host
//   UHS features consequently ignored
//-----------------------------------------

    // CMD2 then CMD3 are required back-to-back.  If the sequence is
    //   unsuccesful then CMD3 will timeout and return a zero value
    send_cmd_r2(2, 0, resp);    // CMD2  ALL_SEND_CID - "ready" to "ident" state, R2 response (CID)
    // also next R1/R6 card status (CMD3) will indicate a partial ACMD state
    if( !send_cmd(3, 0, resp) )    // CMD3  publish new RCA - "ident" to "standby" state, R6 response
        goto faillabel;

    rca = __builtin_bswap16(*(uint16_t *)&resp[1]);
    rca16 = rca;
    rca = rca << 16;
#ifdef SD_DEBUG
    __builtin_printf("Data Transfer Mode entered - Published RCA %08x\n", rca);
#endif

    if( !send_cmd(7, rca, resp) )    // CMD7  SELECT_CARD - "standby" to "tran" state, R1b response
        goto faillabel;

    status = 0;   // card is good to go

    // Need 4-bit bus width for data blocks
    if( !send_acmd(6, 2, resp) )  // ACMD6  SET_BUS_WIDTH - Engage 4-bit parallel data, R1 response
        goto faillabel;
#ifdef SD_DEBUG
    __builtin_printf("4-bit data interface engaged\n");
#endif

    if( CLK_POL ) {
#ifdef SD_DEBUG
        __builtin_printf("Default-Speed access mode\n");
#endif
        goto cmd6passlabel;
    }

    // CMD6  SWITCH_FUNC (Spec 4.3.10, table 4-32):  Select High-Speed Access Mode
    //   This primarily doubles the card's allotted power limit to 200 mA
    //   It also phase shifts the clock-data relationship
    tx_command(6, 0x80_fffff1);    // CMD6  SWITCH_FUNC, Set G1=1 (High Speed)
    if( !rx_response(resp, 6, 2) )    // R1 response, response length, trailing clocks
        goto faillabel;
    // Oddly, CMD6 reads the function table as a data block
    uint8_t  *buff = __builtin_alloca(512);    // data buffer, in hubRAM
    rx_datablocks(buff, 1, _clockfreq()/8UL, resp);    // data length is 64 bytes, CRC will fail
//    _wypin(PIN_CLK, 20);    // SD spec 4.3.10.1, fig 4-14

//    tx_command(19, 0);    // CMD19  SEND_TUNING_BLOCK
//    if( !rx_response(resp, 6, 2) )
//        goto faillabel;
//    rx_datablocks(buff, 1, _clockfreq()/8UL, resp);    // data length is 64 bytes, CRC will fail
        

    // Now to verify
    send_cmd(7, 0, NULL);    // CMD7  DESELECT_CARD - "tran" to "standby" state, no response
    if( !send_cmd_r2(9, rca, resp) )    // CMD9  SEND_CSD - R2 response
        goto faillabel;
    if( resp[4] != 0x5a )    // TRAN_SPEED - SD spec 5.3.3, tables 5-16 and 5-6
        goto faillabel;    // 32h=25MHz, 5Ah=50MHz, 0Bh=100MHz, 2Bh=200MHz
#ifdef SD_DEBUG
    __builtin_printf("High-Speed access mode engaged\n");
#endif

cmd6passlabel:
    send_cmd(7, 0, NULL);    // CMD7  DESELECT_CARD - "tran" to "standby" state, no response

    send_cmd_r2(10, rca, resp);    // CMD10  SEND_CID, R2 response
    memcpy(cid, resp, 17);
#ifdef SD_DEBUG
    __builtin_printf("CID register backed up\n");
#endif
    
    // Time for full speed clocking - After optional High-Speed access mode has been set!
    // Important for timing to stay loose until now because the calibration process
    // phase aligns the rx clock-data relationship - which itself changes between
    // the speed modes.  Calibration would mess up before this point.
    sdcard_clockconf(CLK_DIV);    // set full speed SD clock divider

    if( !calibrate_rxlag(resp, cid) )    // perform rx calibration
        goto faillabel;
#ifdef SD_DEBUG
    __builtin_printf("  SD clock divider %d (%d MHz)\n", CLK_DIV, _clockfreq() / (CLK_DIV * 1_000_000));
/*
    if( send_cmd_r2(10, rca, resp) ) {    // CMD10  SEND_CID, R2 response
        __builtin_printf("CID decode:  ManID=%02X   OEMID=%c%c",resp[1],resp[2],resp[3]);
        __builtin_printf("  Name=%c%c%c%c%c\n",resp[4],resp[5],resp[6],resp[7],resp[8]);
        __builtin_printf("  Ver=%x.%x   Serial=%08X   Date=%d-%d\n", resp[9]>>4, resp[9]&0xf,
                __builtin_bswap32(*(int32_t*)&resp[10]), 2000+((resp[14]<<4|resp[15]>>4)&0xff), resp[15]&0xf );
    }*/
    __builtin_printf("CID decode:  ManID=%02X   OEMID=%c%c",cid[1],cid[2],cid[3]);
    __builtin_printf("  Name=%c%c%c%c%c\n",cid[4],cid[5],cid[6],cid[7],cid[8]);
    __builtin_printf("  Ver=%x.%x   Serial=%08X   Date=%d-%d\n", cid[9]>>4, cid[9]&0xf,
                __builtin_bswap32(*(int32_t*)&cid[10]), 2000+((cid[14]<<4|cid[15]>>4)&0xff), cid[15]&0xf );

    __builtin_printf("SD Card Init Successful\n");
#endif
    Stat = status;

    return status;


faillabel:
#ifdef SD_DEBUG
    __builtin_printf("SD Init Failure!  :-(\n");
#endif
    _wypin(PIN_CLK, 40);
    _waitus(20);
    send_cmd(0, 0, NULL);    // CMD0  GO_IDLE_STATE - no response expected
    _waitus(20);
//    sdcard_power();
    status = STA_NOINIT;
    Stat = status;

    return status;
}



//-----------------------------------------------------------------------
// Read Block(s)
//-----------------------------------------------------------------------
/*
DRESULT disk_read(
    uint8_t drv,       // Physical drive nmuber (0)
    uint8_t *buff,     // Pointer to the data buffer to store read data
    LBA_t firstblk,    // Start block number (LBA)
    LBA_t blocks )     // Block count
{
    uint32_t  cmd, timeout;
    uint8_t  *resp = __builtin_alloca(20);    // response buffer, in hubRAM

    if( disk_status(drv) & STA_NOINIT )
        return RES_NOTRDY;
//    if( !blocks )
//        return RES_PARERR;

    timeout = cnt() + _clockfreq()/8UL;    // 100 ms
    cmd = 0b110<<6 | dat0pin;    // trigger on high level - card ready

    __asm {    // "const" enforces XIP, "volatile" enforces Fcache
		setse2	#0    // cancel triggering before reuse - not needed if POLLSE2 is used
		setse2	cmd
		setq	timeout
		waitse2   wc
	if_nc	mov	cmd, #0    // Not busy
    }

        if( blocks > 1 ) {
            cmd = 18;    //  READ_MULTIPLE_BLOCK
//            __builtin_printf(" RD%x+%x ", firstblk, blocks);
        } else {
            cmd = 17;    //  READ_SINGLE_BLOCK
//            __builtin_printf(" RD%x ", firstblk);
        }

        tx_command(cmd, firstblk);
        if( rx_response(resp, 6, 2) )    // R1 response, response length, trailing clocks
        {
            uint32_t  PIN_LED = ledpin;

            _pinl(PIN_LED);    // activity LED on
            blocks = rx_datablocks(buff, blocks, timeout, resp);
            _pinh(PIN_LED);    // activity LED off
            if( cmd == 18 ) {
                tx_command(12, 0);    // STOP_TRANSMISSION
                rx_response(resp, 6, -500);    // don't much care if this works
            }
        }
    }

    if( !blocks )
        return RES_OK;

errorlabel:
    calibrate_rxlag(resp, cid);
    return RES_ERROR;
}
*/

DRESULT disk_read(
    uint8_t drv,       // Physical drive nmuber (0)
    uint8_t *buff,     // Pointer to the data buffer to store read data
    LBA_t firstblk,    // Start block number (LBA)
    LBA_t blocks )     // Block count
{
    uint8_t  *resp = __builtin_alloca(20);    // response buffer, in hubRAM
    uint32_t  cmd, timeout = _clockfreq()/8UL;    // 100 ms

    if( disk_status(drv) & STA_NOINIT )
        return RES_NOTRDY;
//    if( !blocks )
//        return RES_PARERR;
 
    if( blocks > 1 ) {
        cmd = 18;    //  READ_MULTIPLE_BLOCK
#ifdef SD_DEBUG
        __builtin_printf(" RD%x+%x ", firstblk, blocks);
#endif
    } else {
        cmd = 17;    //  READ_SINGLE_BLOCK
#ifdef SD_DEBUG
        __builtin_printf(" RD%x ", firstblk);
#endif
    }

    if( writing ) {
        writing = 0;
//        expectedblk = -1;
        tx_command(12, 0);    // STOP_TRANSMISSION
        if( !rx_response(resp, 6, -500) )    // R1b response, response length, trailing clocks
            goto errorlabel;    // no response is a write error
    }

    if( !wait_ready(timeout) )    // wait for not busy on DAT0
        goto errorlabel;    // no response is a write error

    tx_command(cmd, firstblk);
    if( !rx_response(resp, 6, 2) )    // R1 response, response length, trailing clocks
        goto errorlabel;    // no response is a write error

    uint32_t  PIN_LED = ledpin;

    _pinl(PIN_LED);    // activity LED on
    blocks = rx_datablocks(buff, blocks, timeout, resp);
    _pinh(PIN_LED);    // activity LED off
    if( cmd == 18 ) {
        tx_command(12, 0);    // STOP_TRANSMISSION
        rx_response(resp, 6, -500);    // don't much care if this works
    }

    if( !blocks )
        return RES_OK;

errorlabel:
    calibrate_rxlag(resp, cid);

    return RES_ERROR;
}



//-----------------------------------------------------------------------
// Write Block(s)
//-----------------------------------------------------------------------

DRESULT disk_write(
    uint8_t drv,            // Physical drive nmuber (0)
    const uint8_t *buff,    // Pointer to the data to be written
    LBA_t firstblk,         // Start block number (LBA)
    LBA_t blocks )          // Block count
{
    uint8_t  *resp = __builtin_alloca(20);    // response buffer, in hubRAM
    uint32_t  PIN_LED = ledpin,
            timeout = _clockfreq()/4UL;    // 250 ms

    if( disk_status(drv) & STA_NOINIT )
        return RES_NOTRDY;
//    if( !blocks )
//        return RES_PARERR;

        if( blocks < 2 ) {
#ifdef SD_DEBUG
            __builtin_printf(" WR%x ", firstblk);
#endif
            if( writing ) {
                writing = 0;
                tx_command(12, 0);    // STOP_TRANSMISSION
                if( !rx_response(resp, 6, -500) )    // R1b response, response length, trailing clocks
                    goto errorlabel;    // no response is a write error
            }

            if( !wait_ready(timeout) )    // wait for not busy on DAT0
                goto errorlabel;    // no response is a write error
            
            tx_command(24, firstblk);    // WRITE_BLOCK
            if( rx_response(resp, 6, 2) ) {    // R1 response, response length, trailing clocks
                _pinl(PIN_LED);    // activity LED on
                if( tx_datablock(buff, timeout) == 0b10_010_1 )
                    blocks = 0;   // CRC matched
                _pinh(PIN_LED);    // activity LED off
            }
        } else {
#ifdef SD_DEBUG
            __builtin_printf(" WR%x+%x=%x ", firstblk, blocks, expectedblk);
#endif
            if( writing && (firstblk != expectedblk) ) {
                writing = 0;
                tx_command(12, 0);    // STOP_TRANSMISSION
                if( !rx_response(resp, 6, -500) )    // R1b response, response length, trailing clocks
                    goto errorlabel;    // no response is a write error
            }

            if( !writing ) {
                if( !wait_ready(timeout) )    // wait for not busy on DAT0
                    goto errorlabel;    // no response is a write error

                tx_command(25, firstblk);    // WRITE_MULTIPLE_BLOCK
                if( !rx_response(resp, 6, 2) )    // R1 response, response length, trailing clocks
                    goto errorlabel;    // no response is a write error

                writing = 1;
            }

            expectedblk = firstblk + blocks;
            _pinl(PIN_LED);    // activity LED on
            do {
                if( tx_datablock(buff, timeout) != 0b10_010_1 )
                    break;   // other than CRC matched
                buff += 512;
            } while( --blocks );
            _pinh(PIN_LED);    // activity LED off
        }

    if( !blocks )
        return RES_OK;

errorlabel:
//    expectedblk = -1;
    if( writing ) {
        writing = 0;
        send_cmd(12, 0, resp);    // STOP_TRANSMISSION
    }
    calibrate_rxlag(resp, cid);

    return RES_ERROR;
}



//-----------------------------------------------------------------------
// Miscellaneous Functions
//-----------------------------------------------------------------------

DRESULT disk_ioctl(
    uint8_t drv,    // Physical drive nmuber (0)
    uint8_t ctrl,    // Control code
    void *buff )    // Buffer to send/receive control data
{
    DRESULT  res;

    if( disk_status(drv) & STA_NOINIT )    // Check if card is in the socket
        return RES_NOTRDY;

    res = RES_ERROR;
    switch( ctrl ) {
        case CTRL_SYNC :    // Make sure that no pending write process
#ifdef SD_DEBUG
            __builtin_printf(" SYNC ");
#endif
//            if( wait_ready(_clockfreq()/4UL) )    // wait for not busy on DAT0, 250 ms timeout
                res = RES_OK;
            break;

        case GET_SECTOR_COUNT :    // Get number of blocks on the disc
        {
            uint32_t  rca = rca16<<16;
            uint8_t  *resp = __builtin_alloca(20);    // response buffer, in hubRAM
            uint8_t  *csd = &resp[1];
            LBA_t  cs;
#ifdef SD_DEBUG
            __builtin_printf(" CAPACITY ");
#endif
            send_cmd(7, 0, NULL);    // DESELECT, no response expected
            if( send_cmd_r2(9, rca, resp) ) {    // SEND_CSD, R2 response
                // SDC ver 2.00
                cs = ((LBA_t)csd[7]<<16 | (LBA_t)csd[8]<<8 | csd[9]) + 1;  // 512 kiByte granularity
                *(LBA_t *)buff = cs << 10;    // 32-bit block count
                res = RES_OK;
            }
            send_cmd(7, rca, resp);    // CMD7  SELECT_CARD - "standby" to "tran" state, R1b response
            break;
        }

        case GET_BLOCK_SIZE :    // Get erase block size in unit of sector
            *(LBA_t *)buff = 128;
            res = RES_OK;
            break;

        default:
            res = RES_PARERR;
    }

    return res;
}



DRESULT disk_setpins( int drv, int CLK_DIV, int PIN_CLK, int PIN_CMD, int PIN_DAT0, int PIN_POW, int PIN_LED )
{
    if( drv )
        return RES_NOTRDY;

    clkdivider = CLK_DIV;
    clkpin = PIN_CLK;
    cmdpin = PIN_CMD;
    dat0pin = PIN_DAT0;
    powpin = PIN_POW;
    ledpin = PIN_LED;
#ifdef SD_DEBUG
    __builtin_printf(" Set pins: CLK_PIN=%d CMD_PIN=%d DAT_PIN=%d POW_PIN=%d LED_PIN=%d\n",
                     PIN_CLK, PIN_CMD, PIN_DAT0, PIN_POW, PIN_LED);
#endif    
    return RES_OK;
}


//
// new routine: deinitialize (clean up pins, if necessary)
//
DSTATUS disk_deinitialize( uint8_t drv )
{
    if( drv ) {
#ifdef SD_DEBUG
        __builtin_printf( " Deinitialize: bad drv %d\n", drv );
#endif
        return RES_NOTRDY;
    }
#ifdef SD_DEBUG
    __builtin_printf( " Clear pins: %d %d %d %d %d\n", clkpin, cmdpin, dat0pin, powpin, ledpin );
#endif
    releasepins();
    _waitms(1);

    return 0;
}
